/****************************************************************************************[Solver.C]
MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include "Solver.h"
#include "Sort.h"
#include <cmath>
#include <signal.h>

#define LIT_TO_VAR(lit) (sign(lit) ? -(var(lit)+1) : var(lit)+1)

//=================================================================================================
// Constructor/Destructor:


Solver::Solver() :

    // Parameters: (formerly in 'SearchParams')
    var_decay(1 / 0.95), clause_decay(1 / 0.999), random_var_freq(0.02)
  , restart_first(100), restart_inc(1.5), restart_last(1000.0)
  , restart_strategy(GEO), restart_type(GLOBAL)
  , restart_inc_last(1.0)
  , max_decision_level(0)
  , learntsize_factor((double)1/(double)3), learntsize_inc(1.1)

   // Activity parameters
  , max_activity(1e100), decrease_activity(1e-100)

    // More parameters:
    //
  , expensive_ccmin  (true)
  , polarity_mode    (polarity_false)
  , verbosity        (0)
  , time_limit		 (INT_MAX-100)

    // Statistics: (formerly in 'SolverStats')
    //
  , starts(0), decisions(0), rnd_decisions(0), propagations(0), conflicts(0)
  , clauses_literals(0), learnts_literals(0), max_literals(0), tot_literals(0)

  , ok               (true)
  , cla_inc          (1)
  , var_inc          (1)
  , qhead            (0)
  , simpDB_assigns   (-1)
  , simpDB_props     (0)
  , order_heap       (VarOrderLt(activity))
  , random_seed      (91648253)
  , remove_satisfied (true)
  , literal_activity (false)
  // dynamic restart variables
  , dynamic_decrement(10.0)
  , dynamic_decrement_mode(ARITHM)
  , minimal_restart_val(10)
  , shrinkingThreshold(95)
  , clsLenSum(0)
  , clsLenSqrSum(0)
  , useShrinking(false)
  , shrLearnClause(true)
  , firstConflictAfterShr(false)
  , clsSizeSum(0)
  , clsDecLevSum(0)
  , shrinkingPerfomred(0)
  , shrinkingReducedSize(0)
  , shrinkingOrder(REVERSE)
  , shrinkingAllClauses(false)
  , shrinkingDecisionsLevels(false)
{
	vecConfl.push(0);
}


Solver::~Solver()
{
    for (int i = 0; i < learnts.size(); i++) free(learnts[i]);
    for (int i = 0; i < clauses.size(); i++) free(clauses[i]);
}


//=================================================================================================
// Minor methods:


// Creates a new SAT variable in the solver. If 'decision_var' is cleared, variable will not be
// used as a decision variable (NOTE! This has effects on the meaning of a SATISFIABLE result).
//
Var Solver::newVar(bool sign, bool dvar)
{
    int v = nVars();
    watches   .push();          // (list for positive literal)
    watches   .push();          // (list for negative literal)
    reason    .push(NULL);
    assigns   .push(toInt(l_Undef));
    level     .push(-1);
    activity  .push(0);
	if (literal_activity)
		activity.push(0);
    seen      .push(0);
    polarity    .push((char)sign);
    decision_var.push((char)dvar);

	if (literal_activity) {
		insertVarOrder(2*v);
		insertVarOrder(2*v+1);
	}
	else 
		insertVarOrder(v);
    return v;
}


bool Solver::addClause(vec<Lit>& ps)
{
    assert(decisionLevel() == 0);

    if (!ok)
        return false;

	// Check if clause is satisfied and remove false/duplicate literals:
    sort(ps);
    Lit p; int i, j;
	for (i = j = 0, p = lit_Undef; i < ps.size(); i++) {
        if (value(ps[i]) == l_True || ps[i] == ~p)
            return true;
        else if (value(ps[i]) != l_False && ps[i] != p)
            ps[j++] = p = ps[i];
	}
    ps.shrink(i - j);

    if (ps.size() == 0)
        return ok = false;
    else if (ps.size() == 1){
        assert(value(ps[0]) == l_Undef);
        uncheckedEnqueue(ps[0]);
        return ok = (propagate() == NULL);
    }else{
        Clause* c = Clause_new(ps, false);
        clauses.push(c);
        attachClause(*c);
    }

    return true;
}


void Solver::attachClause(Clause& c) {
    assert(c.size() > 1);
    watches[toInt(~c[0])].push(&c);
    watches[toInt(~c[1])].push(&c);
    if (c.learnt()) learnts_literals += c.size();
	else clauses_literals += c.size(); 
}


void Solver::detachClause(Clause& c) {
	assert(c.size() > 1);
	assert(find(watches[toInt(~c[0])], &c));
	assert(find(watches[toInt(~c[1])], &c));
	remove(watches[toInt(~c[0])], &c);
	remove(watches[toInt(~c[1])], &c);
	if (c.learnt()) learnts_literals -= c.size();
	else clauses_literals -= c.size();
}


void Solver::removeClause(Clause& c) {
	detachClause(c);
	free(&c); 
}

bool Solver::satisfied(const Clause& c) const {
    for (int i = 0; i < c.size(); i++)
        if (value(c[i]) == l_True)
            return true;
    return false; }


// Revert to the state at given level (keeping all assignment at 'level' but not beyond).
//
void Solver::cancelUntil(int level) {
		// remove levels until "level"
	if (restart_type >= LOCAL) {
			for (int i = decisionLevel(); i > level; i--)
			{
				vecConfl.pop();
			}
		}

	if (decisionLevel() > level){
        for (int c = trail.size()-1; c >= trail_lim[level]; c--){
            Var     x  = var(trail[c]);
            assigns[x] = toInt(l_Undef);
			if (!literal_activity) 
				insertVarOrder(x); 
			else {
				insertVarOrder(2*x); 
				insertVarOrder(2*x+1); 
			}
		}
        qhead = trail_lim[level];
        trail.shrink(trail.size() - trail_lim[level]);
        trail_lim.shrink(trail_lim.size() - level);
    } 
}
//=================================================================================================
// Major methods:


Lit Solver::pickBranchLit(int polarity_mode, double random_var_freq)
{
    Var next = var_Undef;

    // Random decision:
    if (drand(random_seed) < random_var_freq && !order_heap.empty()){
        next = order_heap[irand(random_seed,order_heap.size())];
		if (!literal_activity) {
			if (toLbool(assigns[next]) == l_Undef && decision_var[next])
				rnd_decisions++;
		}
		else {
			if (toLbool(assigns[next/2]) == l_Undef && decision_var[next/2])
				rnd_decisions++;
		}

	}

	if (!literal_activity) { // default behaviour
		// Activity based decision:
		while (next == var_Undef || toLbool(assigns[next]) != l_Undef || !decision_var[next])
			if (order_heap.empty()){
				next = var_Undef;
				break;
			}else
				next = order_heap.removeMin();

		bool sign = false;
		switch (polarity_mode){
		case polarity_true:  sign = false; break;
		case polarity_false: sign = true;  break;
		case polarity_user:  sign = polarity[next]; break;
		case polarity_rnd:   sign = irand(random_seed, 2); break;
		default: assert(false); }

		return next == var_Undef ? lit_Undef : Lit(next, sign);
	}
	else {
		while (next == var_Undef || toLbool(assigns[next/2]) != l_Undef || !decision_var[next/2])
			if (order_heap.empty()){
				next = var_Undef;
				break;
			}else
				next = order_heap.removeMin();
		return next == var_Undef ? lit_Undef : toLit(next);
	}
}


/*_________________________________________________________________________________________________
|
|  analyze : (confl : Clause*) (out_learnt : vec<Lit>&) (out_btlevel : int&)  ->  [void]
|  
|  Description:
|    Analyze conflict and produce a reason clause.
|  
|    Pre-conditions:
|      * 'out_learnt' is assumed to be cleared.
|      * Current decision level must be greater than root level.
|  
|    Post-conditions:
|      * 'out_learnt[0]' is the asserting literal at level 'out_btlevel'.
|  
|  Effect:
|    Will undo part of the trail, upto but not beyond the assumption of the current decision level.
|________________________________________________________________________________________________@*/
void Solver::analyze(Clause* confl, vec<Lit>& out_learnt, int& out_btlevel)
{
    int pathC = 0;
    Lit p     = lit_Undef;

    // Generate conflict clause:
    //
    out_learnt.push();      // (leave room for the asserting literal)
    int index   = trail.size() - 1;
    out_btlevel = 0;

    do{
        assert(confl != NULL);          // (otherwise should be UIP)
        Clause& c = *confl;

        if (c.learnt())
            claBumpActivity(c);

        for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++){
            Lit q = c[j];
			int var_q = var(q);
			
            if (!seen[var_q] && level[var_q] > 0){
				if (literal_activity)
					varBumpActivity(toInt(q));
				else 
					varBumpActivity(var_q);

				seen[var_q] = 1;
                if (level[var_q] >= decisionLevel())
                    pathC++;
                else{
                    out_learnt.push(q);
                    if (level[var(q)] > out_btlevel)
                        out_btlevel = level[var(q)];
                }
            }
        }

        // Select next clause to look at:
        while (!seen[var(trail[index--])]);
        p     = trail[index+1];
        confl = reason[var(p)];
        seen[var(p)] = 0;
        pathC--;

    }while (pathC > 0);
    out_learnt[0] = ~p;

    // Simplify conflict clause:
    //
    int i, j;
    if (expensive_ccmin){
        uint32_t abstract_level = 0;
        for (i = 1; i < out_learnt.size(); i++)
            abstract_level |= abstractLevel(var(out_learnt[i])); // (maintain an abstraction of levels involved in conflict)

        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++)
            if (reason[var(out_learnt[i])] == NULL || !litRedundant(out_learnt[i], abstract_level))
                out_learnt[j++] = out_learnt[i];
    }else{
        out_learnt.copyTo(analyze_toclear);
        for (i = j = 1; i < out_learnt.size(); i++){
            Clause& c = *reason[var(out_learnt[i])];
            for (int k = 1; k < c.size(); k++)
                if (!seen[var(c[k])] && level[var(c[k])] > 0){
                    out_learnt[j++] = out_learnt[i];
                    break; }
        }
    }
    max_literals += out_learnt.size();
    out_learnt.shrink(i - j);
    tot_literals += out_learnt.size();

    // Find correct backtrack level:
    //
    if (out_learnt.size() == 1)
        out_btlevel = 0;
    else{
        int max_i = 1;
        for (int i = 2; i < out_learnt.size(); i++)
            if (level[var(out_learnt[i])] > level[var(out_learnt[max_i])])
                max_i = i;
        Lit p             = out_learnt[max_i];
        out_learnt[max_i] = out_learnt[1];
        out_learnt[1]     = p;
        out_btlevel       = level[var(p)];
    }


    for (int j = 0; j < analyze_toclear.size(); j++) seen[var(analyze_toclear[j])] = 0;    // ('seen[]' is now cleared)
}


// Check if 'p' can be removed. 'abstract_levels' is used to abort early if the algorithm is
// visiting literals at levels that cannot be removed later.
bool Solver::litRedundant(Lit p, uint32_t abstract_levels)
{
    analyze_stack.clear(); analyze_stack.push(p);
    int top = analyze_toclear.size();
    while (analyze_stack.size() > 0){
        assert(reason[var(analyze_stack.last())] != NULL);
        Clause& c = *reason[var(analyze_stack.last())]; analyze_stack.pop();

        for (int i = 1; i < c.size(); i++){
            Lit p  = c[i];
            if (!seen[var(p)] && level[var(p)] > 0){
                if (reason[var(p)] != NULL && (abstractLevel(var(p)) & abstract_levels) != 0){
                    seen[var(p)] = 1;
                    analyze_stack.push(p);
                    analyze_toclear.push(p);
                }else{
                    for (int j = top; j < analyze_toclear.size(); j++)
                        seen[var(analyze_toclear[j])] = 0;
                    analyze_toclear.shrink(analyze_toclear.size() - top);
                    return false;
                }
            }
        }
    }

    return true;
}


/*_________________________________________________________________________________________________
|
|  analyzeFinal : (p : Lit)  ->  [void]
|  
|  Description:
|    Specialized analysis procedure to express the final conflict in terms of assumptions.
|    Calculates the (possibly empty) set of assumptions that led to the assignment of 'p', and
|    stores the result in 'out_conflict'.
|________________________________________________________________________________________________@*/
void Solver::analyzeFinal(Lit p, vec<Lit>& out_conflict)
{
    out_conflict.clear();
    out_conflict.push(p);

    if (decisionLevel() == 0)
        return;

    seen[var(p)] = 1;

    for (int i = trail.size()-1; i >= trail_lim[0]; i--){
        Var x = var(trail[i]);
        if (seen[x]){
            if (reason[x] == NULL){
                assert(level[x] > 0);
                out_conflict.push(~trail[i]);
            }else{
                Clause& c = *reason[x];
                for (int j = 1; j < c.size(); j++)
                    if (level[var(c[j])] > 0)
                        seen[var(c[j])] = 1;
            }
            seen[x] = 0;
        }
    }

    seen[var(p)] = 0;
}


void Solver::uncheckedEnqueue(Lit p, Clause* from)
{
    assert(value(p) == l_Undef);
    assigns [var(p)] = toInt(lbool(!sign(p)));  // <<== abstract but not uttermost effecient
    level   [var(p)] = decisionLevel();
    reason  [var(p)] = from;
    trail.push(p);
}


/*_________________________________________________________________________________________________
|
|  propagate : [void]  ->  [Clause*]
|  
|  Description:
|    Propagates all enqueued facts. If a conflict arises, the conflicting clause is returned,
|    otherwise NULL.
|  
|    Post-conditions:
|      * the propagation queue is empty, even if there was a conflict.
|________________________________________________________________________________________________@*/
Clause* Solver::propagate()
{
    Clause* confl     = NULL;
    int     num_props = 0;

    while (qhead < trail.size()){
        Lit            p   = trail[qhead++];     // 'p' is enqueued fact to propagate.
        vec<Clause*>&  ws  = watches[toInt(p)];
        Clause         **i, **j, **end;
        num_props++;

        for (i = j = (Clause**)ws, end = i + ws.size();  i != end;){
            Clause& c = **i++;

            // Make sure the false literal is data[1]:
            Lit false_lit = ~p;
            if (c[0] == false_lit)
                c[0] = c[1], c[1] = false_lit;

            assert(c[1] == false_lit);

            // If 0th watch is true, then clause is already satisfied.
            Lit first = c[0];
            if (value(first) == l_True){
                *j++ = &c;
            }else{
                // Look for new watch:
                for (int k = 2; k < c.size(); k++)
                    if (value(c[k]) != l_False){
                        c[1] = c[k]; c[k] = false_lit;
                        watches[toInt(~c[1])].push(&c);
                        goto FoundWatch; }

				// Did not find watch -- clause is unit under assignment:
                *j++ = &c;
                if (value(first) == l_False){
                   confl = &c;
                   qhead = trail.size();
                   // Copy the remaining watches:
                   while (i < end)
					*j++ = *i++;
				}
				else {
					uncheckedEnqueue(first, &c);
				}
			}
        FoundWatch:;
        }
        ws.shrink(i - j);
    }
    propagations += num_props;
    simpDB_props -= num_props;
    return confl;
}

/*_________________________________________________________________________________________________
|
|  reduceDB : ()  ->  [void]
|  
|  Description:
|    Remove half of the learnt clauses, minus the clauses locked by the current assignment. Locked
|    clauses are clauses that are reason to some assignment. Binary clauses are never removed.
|________________________________________________________________________________________________@*/
struct reduceDB_lt { bool operator () (Clause* x, Clause* y) { return x->size() > 2 && (y->size() == 2 || x->activity() < y->activity()); } };
void Solver::reduceDB()
{
    int     i, j;
    double  extra_lim = cla_inc / learnts.size();    // Remove any clause below this activity

    sort(learnts, reduceDB_lt());
    for (i = j = 0; i < learnts.size() / 2; i++){
        if (learnts[i]->size() > 2 && !locked(*learnts[i]))
            removeClause(*learnts[i]);
        else
            learnts[j++] = learnts[i];
    }
    for (; i < learnts.size(); i++){
        if (learnts[i]->size() > 2 && !locked(*learnts[i]) && learnts[i]->activity() < extra_lim)
            removeClause(*learnts[i]);
        else
            learnts[j++] = learnts[i];
    }
	learnts.shrink(i-j);
}

void Solver::removeSatisfied(vec<Clause*>& cs)
{
    int i,j;
    for (i = j = 0; i < cs.size(); i++){
        if (satisfied(*cs[i]))
            removeClause(*cs[i]);
        else
            cs[j++] = cs[i];
    }
    cs.shrink(i - j);
}


/*_________________________________________________________________________________________________
|
|  simplify : [void]  ->  [bool]
|  
|  Description:
|    Simplify the clause database according to the current top-level assigment. Currently, the only
|    thing done here is the removal of satisfied clauses, but more things can be put here.
|________________________________________________________________________________________________@*/
bool Solver::simplify()
{
    assert(decisionLevel() == 0);

    if (!ok || propagate() != NULL)
        return ok = false;

    if (nAssigns() == simpDB_assigns || (simpDB_props > 0))
        return true;

    // Remove satisfied clauses:
    removeSatisfied(learnts);
    if (remove_satisfied)        // Can be turned off.
        removeSatisfied(clauses);

    // Remove fixed variables from the variable heap:
    order_heap.filter(VarFilter(*this));

    simpDB_assigns = nAssigns();
    simpDB_props   = clauses_literals + learnts_literals;   // (shouldn't depend on stats really, but it will do for now)

    return true;
}

inline double Solver::stddev(
	int n,
	double sqrsum,
	double sum)
{
	return sqrt(((double)n * sqrsum - sum * sum) / ((double)n * (double)n));
}

inline double Solver::mean(
	int n,
	double sum)
{
	return (double)sum / (double)n;
}

inline void Solver::calculateShrinkingThresold(int clsNum)
{
	// shrinking update 
	double clsMean = mean(clsNum, clsLenSum);
	double clsStdDev  = stddev(clsNum, clsLenSqrSum, clsLenSum);
	double center = clsMean + 0.5*clsStdDev;
	double ulimit = clsMean + clsStdDev;

	if (shrinkingThreshold >= center)
		shrinkingThreshold -= 5;
	if (shrinkingThreshold < center)
		shrinkingThreshold += 5;
	if (shrinkingThreshold > ulimit)
		shrinkingThreshold = ulimit;
	if (shrinkingThreshold < 5)
		shrinkingThreshold = 5;

	clsLenSum = 0;
	clsLenSqrSum = 0;
}

double Solver::getValue(Var v, ShrinkingOrder shrOrd)
{
	switch (shrOrd)
	{
		case SAME: return -level[v];
		case REVERSE: return level[v];
		case ACTIVITY: return activity[v];
	}
	
	return 0.0;
}

void Solver::sortClauseByLevels(vec<Lit>& cls, ShrinkingOrder shrOrd)
{
	double largest, median;
	double val[3];
	int i, j, lo, hi;
	int top;
	int stack[60];  /* Suffices for up to 2^30 elements */
	stack[0] = 0;
	stack[1] = cls.size() - 1;
	top = 2;
	do
	{
		top -= 2;
		lo = stack[top];
		hi = stack[top + 1];
		if (lo < hi)
		{
			if (hi < lo + 12)      /* Selection sort for small arrays */
			{
				int min_index;
				Lit temp;
				for (i = lo; i < hi; i++)
				{
					min_index = i;
					for (j = i + 1; j <= hi; j++)
					{
						if (getValue(var(cls[j]), shrOrd) < getValue(var(cls[min_index]), shrOrd))
						{
							min_index = j;
						}
					}

					temp = cls[i];
					cls[i] = cls[min_index];
					cls[min_index] = temp;
				}
			}
			else    /* Quicksort for large arrays */
			{
				/*  Find a good partition element using the median-of-three
				method  */
				val[0] = getValue(var(cls[lo]), shrOrd);
				val[1] = getValue(var(cls[(lo + hi) / 2]), shrOrd);
				val[2] = getValue(var(cls[hi]), shrOrd);
				largest = val[0];
				if (val[1] >= val[0])
				{
					median = val[0];
					largest = val[1];
				}
				else
				{
					median = val[1];
				}

				if (val[2] >= largest)
				{
					median = largest;
				}
				else if (val[2] > median)
				{
					median = val[2];
				}

				/*  Partition the array  */
				i = lo;
				j = hi;
				do
				{
					while ((getValue(var(cls[i]), shrOrd) < median) && (i <= hi))
					{
						i++;
					}

					while ((getValue(var(cls[j]), shrOrd) > median) && (j >= lo))
					{
						j--;
					}

					if (i <= j)
					{
						Lit temp = cls[i];
						cls[i] = cls[j];
						cls[j] = temp;
						i++;
						j--;
					}
				} while (i <= j);

				/*  Put the largest chunk on the stack and process
				the smaller immediately  */
				if ((i - lo) > (hi - i))
				{
					stack[top] = lo;
					stack[top + 1] = j;
					stack[top + 2] = i;
					stack[top + 3] = hi;
				}
				else
				{
					stack[top] = i;
					stack[top + 1] = hi;
					stack[top + 2] = lo;
					stack[top + 3] = j;
				}

				top += 4;
			}
		}
	} while (top > 0);
}


int Solver::calculateDecisionLevels(vec<Lit>& cls)
{
	int decLevels = 0;
	for (int i=0; i<cls.size(); ++i )
	{
		int levelV = level[var(cls[i])];
		if (decLevInConfl[levelV] != conflicts)
		{
			++decLevels;
			decLevInConfl[levelV] = conflicts;
		}
	}

	return decLevels;
}

/*_________________________________________________________________________________________________
|
|  search : (nof_conflicts : int) (nof_learnts : int) (params : const SearchParams&)  ->  [lbool]
|  
|  Description:
|    Search for a model the specified number of conflicts, keeping the number of learnt clauses
|    below the provided limit. NOTE! Use negative value for 'nof_conflicts' or 'nof_learnts' to
|    indicate infinity.
|  
|  Output:
|    'l_True' if a partial assigment that is consistent with respect to the clauseset is found. If
|    all variables are decision variables, this means that the clause set is satisfiable. 'l_False'
|    if the clause set is unsatisfiable. 'l_Undef' if the bound on number of conflicts is reached.
|________________________________________________________________________________________________@*/
lbool Solver::search(int nof_conflicts, int nof_learnts)
{
    assert(ok);
    int         backtrack_level;
    int         conflictC = 0;
    vec<Lit>    learnt_clause;

    starts++;

	while(true) {
        Clause* confl = propagate();
        if (confl != NULL){
            // CONFLICT
			conflicts++; conflictC++;
            if (decisionLevel() == 0)
					return l_False;

            learnt_clause.clear();
            analyze(confl, learnt_clause, backtrack_level);
			long long clsSize = learnt_clause.size();
			clsSizeSum += clsSize;
			
			if (useShrinking)
			{
				long long clsDecLevSize = calculateDecisionLevels(learnt_clause);
				clsDecLevSum += clsDecLevSize;
				if (!shrinkingDecisionsLevels)
				{
					clsLenSum += clsSize;
					clsLenSqrSum += clsSize*clsSize;
				}
				else
				{
					clsLenSum += clsDecLevSize;
					clsLenSqrSum += clsDecLevSize*clsDecLevSize;
				}

				if (shrinkingCls.size() > 0)
				{
					shrinkingCls.clear();
					shrinkingReducedSize++;
				}

				if (!shrinkingAllClauses && firstConflictAfterShr) 
				{
					firstConflictAfterShr = false;
				}
				else 
				{
					if ((!shrinkingDecisionsLevels && clsSize > shrinkingThreshold) ||
						(shrinkingDecisionsLevels && clsDecLevSize > shrinkingThreshold))
					{
						// check if there is a backtrack option
						learnt_clause.copyTo(shrinkingCls);
						sortClauseByLevels(shrinkingCls, shrinkingOrder);

						// find the backtrack level
						int size = shrinkingCls.size();

						switch (shrinkingOrder)
						{
						case REVERSE: 
							for (int i=1; i < size; ++i)
							{
								if (level[var(shrinkingCls[i])] > level[var(shrinkingCls[i-1])]+1)
								{
									int bckLevel = level[var(shrinkingCls[i-1])];
									if (bckLevel < backtrack_level) 
									{
										cancelUntil(bckLevel);
										break;
									}
								}
							}
							break;
						case SAME:
							for (int i=shrinkingCls.size()-2; i >= 0; --i)
							{
								if (level[var(shrinkingCls[i])] > level[var(shrinkingCls[i+1])]+1)
								{
									int bckLevel = level[var(shrinkingCls[i+1])];
									if (bckLevel < backtrack_level) 
									{
										cancelUntil(bckLevel);
										break;
									}
								}
							}
							break;
						case ACTIVITY:
							vec<Lit> orderedCls;
							shrinkingCls.copyTo(orderedCls);

							sortClauseByLevels(orderedCls, REVERSE);
							for (int i=1; i < size; ++i)
							{
								if (level[var(orderedCls[i])] > level[var(orderedCls[i-1])]+1)
								{
									int bckLevel = level[var(orderedCls[i-1])];
									if (bckLevel < backtrack_level) 
									{
										cancelUntil(bckLevel);
										break;
									}
								}
							}
							orderedCls.clear();
							break;
						}

						if (decisionLevel() < backtrack_level)
						{
							shrinkingPerfomred++;
							if (shrLearnClause)
							{
								Clause* c;
								while ((c= Clause_new(learnt_clause, true)) == NULL) {
									reduceDB();
								}
								learnts.push(c);
								attachClause(*c);
								claBumpActivity(*c);
							}
							firstConflictAfterShr = true;
							continue;
						}
						else
						{
							shrinkingCls.clear();
						}
					}
				}

				if (conflicts%2000 == 0) 
				{
					calculateShrinkingThresold(2000);
				}
			}

			cancelUntil(backtrack_level);
			assert(value(learnt_clause[0]) == l_Undef);

			if (learnt_clause.size() == 1){
                uncheckedEnqueue(learnt_clause[0]);
            }else{
				Clause* c;
				while ((c= Clause_new(learnt_clause, true)) == NULL) {
					reduceDB();
				}
                learnts.push(c);
				attachClause(*c);
		        claBumpActivity(*c);
                uncheckedEnqueue(learnt_clause[0], c);
            }

            varDecayActivity();
            claDecayActivity();

        }else{
			if (cpuTime() > time_limit) {
				raise(SIGINT);
			}

            // NO CONFLICT
			int dLevel = decisionLevel();
			if ((dLevel > 0) && (
				((restart_type == GLOBAL) && (nof_conflicts >= 0 && conflictC >= nof_conflicts)) ||
				((restart_type == LOCAL) && (dLevel > 0) && (conflictC - vecConfl[dLevel] >= nof_conflicts)) ||
				((restart_type == DYNAMIC) && (dLevel > 0) && 
					((dynamic_decrement_mode == ARITHM) && 
						(conflictC - vecConfl[dLevel] >= nof_conflicts - dynamic_decrement*(dLevel-1)) && (conflictC - vecConfl[dLevel] > minimal_restart_val)) ||
					((dynamic_decrement_mode == GEO) && 
						(conflictC - vecConfl[dLevel] >= nof_conflicts/(dynamic_decrement*dLevel)) && (conflictC - vecConfl[dLevel] > minimal_restart_val))
				) ||
				((restart_type == PROGRESSIVE_DYNAMIC) && (conflictC - vecConfl[dLevel] >= calculateProgressiveRestartBound(nof_conflicts)))
			   ))

			{
                // Reached bound on number of conflicts:
                cancelUntil(0);
                return l_Undef; 
			}

            // Simplify the set of problem clauses:
            if (decisionLevel() == 0 && !simplify())
                return l_False;

            if (nof_learnts >= 0 && learnts.size()-nAssigns() >= nof_learnts)
                // Reduce the set of learnt clauses:
                reduceDB();

            Lit next = lit_Undef;
			if (useShrinking && shrinkingCls.size() > 0)
			{
				while ((next == lit_Undef || value(next) != l_Undef) && (shrinkingCls.size() > 0))
				{
					next = ~shrinkingCls[shrinkingCls.size()-1];
					shrinkingCls.pop();
				}
				if (shrinkingCls.size() == 0)
					next = lit_Undef;
			}

			if (next == lit_Undef) 
			while (decisionLevel() < assumptions.size()){
                // Perform user provided assumption:
                Lit p = assumptions[decisionLevel()];
                if (value(p) == l_True){
                    // Dummy decision level:
                    newDecisionLevel();
                }else if (value(p) == l_False){
                    analyzeFinal(~p, conflict);
                    return l_False;
                }else{
                    next = p;
                    break;
                }
            }

            if (next == lit_Undef){
                // New variable decision:
                decisions++;
                next = pickBranchLit(polarity_mode, random_var_freq);

                if (next == lit_Undef)
                    // Model found:
                    return l_True;
            }

            // Increase decision level and enqueue 'next'
            assert(value(next) == l_Undef);
            newDecisionLevel();
			if (restart_type >= LOCAL)
				vecConfl.push(conflictC);

            uncheckedEnqueue(next);
        }
    }
}


bool Solver::solve(const vec<Lit>& assumps)
{
	struct Luby luby;

	model.clear();
    conflict.clear();

	if (!ok) return false;

	decLevInConfl.growTo(nVars(), 0);

    assumps.copyTo(assumptions);

	double  nof_conflicts = restart_first;
	if (restart_strategy == LUBY)
		nof_conflicts = luby.next()*restart_first;

    double  nof_learnts   = nClauses() * learntsize_factor;
    lbool   status        = l_Undef;

    if (verbosity >= 1){
        reportf("============================[ Search Statistics ]===================\n");
        reportf("| Conflicts |          ORIGINAL         |          LEARNT          |\n");
        reportf("|           |    Vars  Clauses Literals |    Limit  Clauses Lit/Cl |\n");
        reportf("====================================================================\n");
    }

    // Search:
    while (status == l_Undef){
		if (verbosity >= 1)
            reportf("| %9d | %7d %8d %8d | %8d %8d %6.0f |\n", (int)conflicts, order_heap.size(), nClauses(), (int)clauses_literals, (int)nof_learnts, nLearnts(), (double)learnts_literals/nLearnts()), fflush(stdout);
		status = search((int)ceil(nof_conflicts), (int)nof_learnts);
        
		switch (restart_strategy) {
			case ARITHM:
				nof_conflicts += restart_inc;
				break;
			case GEO:
				nof_conflicts *= restart_inc;
				break;
			case LUBY:
				nof_conflicts = luby.next()*restart_first;
				break;
			case ARMIN:
				if (nof_conflicts >= restart_last) {
					restart_last *= restart_inc_last;
					nof_conflicts = restart_first;
				}
				else {
					nof_conflicts *= restart_inc;
				}
				break;
		}

		nof_learnts   *= learntsize_inc;
    }

    if (verbosity >= 1)
        reportf("===============================================================================\n");


    if (status == l_True){
        // Extend & copy model:
        model.growTo(nVars());
        for (int i = 0; i < nVars(); i++) model[i] = value(i);
#ifndef NDEBUG
        verifyModel();
#endif
    }else{
        assert(status == l_False);
        if (conflict.size() == 0)
            ok = false;
    }

    cancelUntil(0);
    return status == l_True;
}

//=================================================================================================
// Debug methods:
void Solver::print_all()
{
	FILE* fout=fopen("c:\\results\\minisat.txt", "w");
	fprintf(fout, "Original Clauses\n");
	for (int i=0; i < clauses.size(); i++) {
		Clause& c = *clauses[i];
		for (int j=0; j<c.size(); j++) {
			fprintf(fout, "%d ", LIT_TO_VAR(c[j]));
		}
		fprintf(fout, "\n");
	}
	fprintf(fout, "Learnt Clauses\n");
	for (int i=0; i < learnts.size(); i++) {
		Clause& c = *learnts[i];
		for (int j=0; j<c.size(); j++) {
			fprintf(fout, "%d ", LIT_TO_VAR(c[j]));
		}
		fprintf(fout, "\n");
	}
	fclose(fout);

}


void Solver::verifyModel()
{
    bool failed = false;
    for (int i = 0; i < clauses.size(); i++){
        assert(clauses[i]->mark() == 0);
        Clause& c = *clauses[i];
        for (int j = 0; j < c.size(); j++)
            if (modelValue(c[j]) == l_True)
                goto next;

        reportf("unsatisfied clause: ");
        printClause(*clauses[i]);
        reportf("\n");
        failed = true;
    next:;
    }

    assert(!failed);

    reportf("Verified %d original clauses.\n", clauses.size());
}


void Solver::checkLiteralCount()
{
    // Check that sizes are calculated correctly:
    int cnt = 0;
    for (int i = 0; i < clauses.size(); i++)
        if (clauses[i]->mark() == 0)
            cnt += clauses[i]->size();

    if ((int)clauses_literals != cnt){
        fprintf(stderr, "literal count: %d, real value = %d\n", (int)clauses_literals, cnt);
        assert((int)clauses_literals == cnt);
    }
}

int Solver::calculateProgressiveRestartBound(int nof_conflicts)
{
	// do nothing before first restart
	if (starts < 2) {
		return nof_conflicts;
	}
	int dl = decisionLevel();

	if (max_decision_level == 0) {
		return nof_conflicts;
	}

	int middle = max_decision_level*dynamic_decrement;
	double part = (double)nof_conflicts / (max_decision_level == middle ? max_decision_level : abs(max_decision_level - middle));
	int result = (int)(nof_conflicts+part*(middle-dl));

	return result > minimal_restart_val ? result : minimal_restart_val;
}